"""
QMS scenario generator v1.0. Copyright Unigraf 2024

This is a tool to analyse video recordings of QMS test pattern.
"""

import cv2 as cv
import numpy
from matplotlib import pyplot as plt
import time
import sys
import os
import shutil

version = 'v1.0'

# The same image size is used for ROI selection and in processing where ROI is applied
preview_size = (1000, 800)

# Show cumulative histograms of the first 20 frames in advanced mode. Change to True to enable.
histogram_show = False
histogram_previews = 20

# Folder for failed frames
failed_dir = 'failed'


def find_step(steps) -> int:
    # we expect not more than 3 bright steps
    count = sum(steps)
    if count < 1 or count > 3:
        return -2

    number = len(steps)
    if steps[number - 1] and steps[0]:
        for index, step in enumerate(steps):
            if not step:
                return index - 1
    else:
        last = -1
        for index, step in enumerate(steps):
            if step:
                last = index
        return last

    return -3  # not expected


def save_failed(position, frame_number, previous, frame, thresh, thresh0, full, full0):
    if previous is not None:
        cv.imwrite(f'{failed_dir}{os.path.sep}{position}_{frame_number:04}before.jpg', previous)
    if frame is not None:
        cv.imwrite(f'{failed_dir}{os.path.sep}{position}_{frame_number:04}fail.jpg', frame)
    if thresh is not None:
        cv.imwrite(f'{failed_dir}{os.path.sep}{position}_{frame_number:04}thresh.jpg', thresh)
    if thresh0 is not None:
        cv.imwrite(f'{failed_dir}{os.path.sep}{position}_{frame_number:04}thresh0.jpg', thresh0)
    if full is not None:
        cv.imwrite(f'{failed_dir}{os.path.sep}{frame_number:04}fail.jpg', full)
    if full0 is not None:
        cv.imwrite(f'{failed_dir}{os.path.sep}{frame_number:04}before.jpg', full0)


def analyse(name, roi, position, number, side, window, increment, level):
    cap = cv.VideoCapture(name)
    step = -1
    expected = window * window // 2
    previous_full = None
    previous = None
    thresh0 = None
    frame_number = 0
    errors = 0
    header = f'\n===== Analysing {name} at {position} pattern row.'
    print(header)
    log.write(header + '\n')

    while cap.isOpened():
        ret, frame0 = cap.read()
        if ret:
            frame = cv.resize(frame0, preview_size)

            # Crop image
            cropped = frame[int(roi[1]):int(roi[1] + roi[3]),
                            int(roi[0]):int(roi[0] + roi[2])]
            if not len(cropped):
                break

            cropped = cv.resize(cropped, (number * side, side))

            img = cv.cvtColor(cropped, cv.COLOR_BGR2GRAY)
#            img = cv.equalizeHist(img)

            # applying Otsu thresholding as an extra flag in binary thresholding
            if level == 0:
                ret, thresh1 = cv.threshold(img, 200, 255, cv.THRESH_BINARY + cv.THRESH_OTSU)
            else:
                ret, thresh1 = cv.threshold(img, level, 255, cv.THRESH_BINARY)

            info = ''
            left = upper = (side - window) // 2
            right = lower = (side + window) // 2
            counts = []
            for index in range(number):
                part = thresh1[upper:lower, left:right]
                count = cv.countNonZero(part)
                info += f'{count:5} '
                counts.append(1 if count > expected else 0)
                cv.rectangle(cropped, (left, upper), (right, lower), (255, 255, 0), 2)
                left += side
                right += side

            now = find_step(counts)
            cv.imshow('Unigraf QMS Pattern Analyser', cropped)

            print('\r', f'{frame_number:5}:{info}:{now}', end='')

            expected = step + increment
            if expected < 0:
                expected = number - 1
            elif expected >= number:
                expected = 0

            if now < 0:
                save_failed(position, frame_number, previous, cropped, thresh1, thresh0, frame, previous_full)
                errors += 1
                print(f' pattern not recognized. errors:{errors}')
                log.write(f'\nframe {frame_number:5} pattern not recognized. errors:{errors}')
            elif step < 0:
                step = now
            elif now == step:
                # no changes
                pass
            elif now != expected:
                save_failed(position, frame_number, previous, cropped, thresh1, thresh0, frame, previous_full)
                errors += 1
                print(f' {step} -> {now} expected {expected} errors:{errors}')
                log.write(f'\nframe {frame_number:5} {step} -> {now} expected {expected}. errors:{errors}')
                step = expected
            else:
                step = now

            frame_number += 1
            previous_full = frame
            previous = cropped
            thresh0 = thresh1

            if errors >= 50:
                print('Too many failures')
                break

            if cv.waitKey(1) == ord('q'):
                break
        else:
            break

    report = f'\nScanned frames: {frame_number} errors: {errors}'
    print(report)
    log.write(f'{report}\n')

    cap.release()
    cv.destroyAllWindows()


def show_histogram(name, position, roi):
    cap = cv.VideoCapture(name)
    if histogram_show:
        print(f'\n===== Building histogram of {name} at {position} pattern row.')
    frame_number = 0
    color = ('b', 'g', 'r')
    accumulated = dict()

    while cap.isOpened():
        ret, frame0 = cap.read()
        if ret:
            frame = cv.resize(frame0, preview_size)
            frame_number += 1

            if not histogram_show or frame_number > histogram_previews:
                break

            print('\r', frame_number, end='')

            # Crop image
            cropped = frame[int(roi[1]):int(roi[1] + roi[3]),
                            int(roi[0]):int(roi[0] + roi[2])]
            if not len(cropped):
                break

            # grey = cv.equalizeHist(grey)
            for i, col in enumerate(color):
                histr = cv.calcHist([cropped], [i], None, [256], [0, 256])
                if col in accumulated.keys():
                    accumulated[col] += histr
                else:
                    accumulated[col] = histr
        else:
            break

    if histogram_show:
        if len(accumulated):
            print('\nCheck the histogram. Decide if fixed threshold value should be used.')
            for i, col in enumerate(color):
                plt.plot(accumulated[col], color=col)
                plt.xlim([0, 256])
            plt.ioff()
            plt.show()
        else:
            print('\nHistogram is not available.')

    return roi


def analyse_all_parts(name):

    full, frame = select_frame_area(name)
    started = time.time()

    roi1 = None
    roi2 = None

    if full is not None and frame is not None:
        height = full[3] // 4
        gap = height // 4
        roi1 = (full[0], full[1] + gap, full[2], height - gap * 2)
        roi2 = (full[0], full[1] + height * 2 + gap, full[2], height - gap * 2)
        cv.rectangle(frame, (full[0], full[1]), (full[0] + full[2], full[1] + full[3]), (0, 128, 255), 3)
        cv.rectangle(frame, (roi1[0], roi1[1]), (roi1[0] + roi1[2], roi1[1] + roi1[3]), (255, 128, 0), 5)
        cv.rectangle(frame, (roi2[0], roi2[1]), (roi2[0] + roi2[2], roi2[1] + roi2[3]), (255, 128, 0), 5)
        cv.imshow('Selected ROI areas', frame)
        cv.imwrite('roi_selected.jpg', frame)
        cv.waitKey(2000)
        cv.destroyAllWindows()

    roi = show_histogram(name, '1st', roi1)
    if len(roi) and roi[1]:
        if histogram_show:
            level = int(input('Fixed threshold level [1..255], or 0 for automatic '))
        else:
            level = 0
        analyse(name, roi,'1st', number=6, side=100, window=10, increment=1, level=level)

    roi = show_histogram(name, '3rd', roi2)
    if len(roi) and roi[1]:
        if histogram_show:
            level = int(input('Fixed threshold level [1..255], or 0 for automatic '))
        else:
            level = 0
        analyse(name, roi,'3rd', number=12, side=50, window=10, increment=-1, level=level)

    return started


def clear_failed_directory():
    try:
        if os.path.isfile(failed_dir) or os.path.islink(failed_dir):
            os.unlink(failed_dir)
        elif os.path.isdir(failed_dir):
            shutil.rmtree(failed_dir)
    except Exception as e:
        print('Failed to delete %s. Reason: %s' % (failed_dir, e))
        return False

    try:
        os.mkdir(failed_dir)
    except OSError:
        return False

    return True


"""
# The function uses selectROIs to select multiple regions. The interface doesn't look clear - the order of actions
# is too complicated - select region, press Enter, select another region, press Enter, and then press ESC to complete
# the selection.
# It's not used now.
def select_rois(name):
    cap = cv.VideoCapture(name)
    print(f'\n===== Selecting ROIs for {name}.')

    if cap.isOpened():
        ret, frame0 = cap.read()
        if ret:
            frame = cv.resize(frame0, preview_size)

            print(f'Select the 1st and the 3rd rows of the test pattern.'
                  '\nPress SPACE or ENTER button to confirm selection.')
            print('Cancel the selection process by pressing ESC or c button.')
            rois = cv.selectROIs(f'Select the row of the test pattern', frame,
                                 printNotice= False, showCrosshair=False)
            cv.destroyAllWindows()

            return rois
    return None
"""


def select_frame_area(name):
    cap = cv.VideoCapture(name)
    frame_number = 0
    accumulated = None

    while cap.isOpened():
        ret, frame0 = cap.read()
        if ret:
            frame = cv.resize(frame0, preview_size)
            if accumulated is None:
                accumulated = frame
            else:
                accumulated = numpy.maximum(accumulated, frame)
        else:
            break

        frame_number += 1
        if frame_number > 60:
            break

    if accumulated is not None:
        print(f'Select the frame area and press SPACE or ENTER button.')
        print('Cancel the selection process by pressing ESC or c button.')
        roi = cv.selectROI(f'Select the frame area', accumulated,
                           printNotice=False, showCrosshair=False)
        cv.destroyAllWindows()
        return roi, accumulated

    return None, None


if __name__ == '__main__':
    caption = f'QMS Video Validation Tool {version}. Copyright Unigraf 2024\n'
    print(caption)

    log = open('analyzer.log', 'w')
    log.write(f'{caption}\n')

    if len(sys.argv) == 1:
        directory = 'samples'
        file_name = 'test_good.mp4'
        path = f'{directory}{os.path.sep}{file_name}'
    else:
        path = sys.argv[1]

    if not clear_failed_directory():
        exit(1)

    log.write(f'{"Advanced" if histogram_show else "Basic"} mode selected.\n')

    started = analyse_all_parts(path)

    elapsed = f'Done in {time.time() - started:.1f} seconds.'
    log.write(f'{elapsed}\n')

    input(f'{elapsed} Press ENTER.')
